import socket
import logging
import re
import itertools
from urllib.parse import urlparse
from typing import *
from .scpi_transport_base import ScpiTransportBase
from .scpi_resource import ScpiResource
from .scpi_exceptions import ScpiTransportException


class ScpiTcpIpTransport(ScpiTransportBase):
    DEFAULT_PORT: int = 5025

    _transport_class = 'ScpiTcpIpTransport'
    _transport_info = 'TCP/IP SCPI Transport'
    _transport_type = 'TCP/IP'

    _sock: Optional[socket.socket] = None

    @classmethod
    def discover(cls,
                 dnssd_services: List[str] = ('_scpi-raw._tcp.local.',),
                 dnssd_domains: List[str] = ('local',),
                 dnssd_names: List[str] = ('.*',),
                 dnssd_ipversions: str = '',
                 dnssd_timeout: float = 0.5) -> List[ScpiResource]:
        from zeroconf import Zeroconf, ServiceListener, ServiceBrowser, IPVersion
        import time

        class Listener(ServiceListener):
            def __init__(self):
                self.services = {}

            def remove_service(self, zeroconf, zc_type, zc_name):
                self.services.pop(zc_name)

            def add_service(self, zeroconf, zc_type, zc_name):
                self.services.update({zc_name: zeroconf.get_service_info(zc_type, zc_name)})


            def update_service(self, zeroconf, zc_type, zc_name):
                self.services.update({zc_name: zeroconf.get_service_info(zc_type, zc_name)})

        # Generate fully qualified service names by permuting given services with given domains
        dnssd_services_fq = [f'{s}.{d}' + ('.' if not d.endswith('.') else '')
                             for s, d in itertools.product(dnssd_services, dnssd_domains)]

        # Find devices via zeroconf mDNS
        # TODO: Implement DNS-SD for non-.local domains
        ip_version = IPVersion.All if '4' in dnssd_ipversions and '6' in dnssd_ipversions else \
            IPVersion.V4Only if '4' in dnssd_ipversions else IPVersion.V6Only if '6' in dnssd_ipversions else None

        listener = Listener()
        zc = Zeroconf(ip_version=ip_version)
        ServiceBrowser(zc, dnssd_services_fq, listener=listener)

        # Wait for some time to get answers
        time.sleep(dnssd_timeout)

        # Patterns to check name against
        patterns = [re.compile(pattern) for pattern in dnssd_names]

        return [ScpiResource(transport=ScpiTcpIpTransport,
                             location=f'dnssd:{service.name}',
                             address=(parsed_address, int(service.port)),
                             name=service.get_name(),
                             manufacturer=service.properties[b'Manufacturer'].decode(
                                 'utf-8') if b'Manufacturer' in service.properties else None,
                             model=service.properties[b'Model'].decode(
                                 'utf-8') if b'Model' in service.properties else None,
                             serialnum=service.properties[b'SerialNumber'].decode(
                                 'utf-8') if b'SerialNumber' in service.properties else None,
                             info=service
                             ) for service in listener.services.values() for parsed_address in service.parsed_scoped_addresses()
                if any([pattern.fullmatch(service.get_name()) for pattern in patterns])]

    @classmethod
    def from_resource_name(cls, resource_name: str) -> Optional[ScpiResource]:
        m = re.match((
            r'^(?P<prefix>(?P<type>TCPIP)(?P<board>\d+)?)'
            r'::((?P<host_address>[^:]+)|(\[(?P<host_address6>.+)\]))'
            r'::(?P<host_port>([0-9]+))'
            r'(::(?P<suffix>SOCKET))?$'
        ), resource_name, re.IGNORECASE)

        if m is None:
            # Does not match the regex
            return None

        groupdict = m.groupdict()

        # Get either IPv4 or IPv6 address
        host_address = groupdict['host_address'] if groupdict['host_address'] is not None else groupdict['host_address6']
        host_port = groupdict['host_port']

        return ScpiResource(
            transport=ScpiTcpIpTransport,
            address=(host_address, host_port)
        )

    @classmethod
    def to_resource_name(cls, resource: ScpiResource) -> str:
        host_address = resource.address[0] if not ':' in resource.address[0] else f'[{resource.address[0]}]'
        host_port = resource.address[1] if len(resource.address) > 1 else cls.DEFAULT_PORT

        return f'TCPIP::{host_address}::{host_port}::SOCKET'

    def __init__(self, address: Union[tuple, str], timeout: float=5.0, **kwargs):
        super().__init__(**kwargs)
        self._logger = logging.getLogger(__name__)
        self._sock = None

        if isinstance(address, str):
            # Convert from string to tuple first
            addr = urlparse('//' + address)
            if (addr.hostname is None):
                raise ScpiTransportException(f'{address} is not a valid hostname')
            address = (addr.hostname, addr.port) if addr.port is not None else (addr.hostname, self.DEFAULT_PORT)

        try:
            # Get address info
            addr_infos = socket.getaddrinfo(address[0], address[1], type=socket.SOCK_STREAM, proto=socket.IPPROTO_TCP)
        except socket.gaierror as e:
            raise ScpiTransportException(e) from e

        for addr_info in addr_infos:
            try:
                # Try to connect
                self._sock = socket.socket(addr_info[0], addr_info[1], addr_info[2])
                self._sock.settimeout(timeout)
                self._sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
                self._sock.connect(addr_info[4])

                # Make a file interface for easier handling
                self._io = self._sock.makefile('rwb')
                self._io.flush()

                # Successful, do not try other addresses
                break
            except OSError as e:
                self._logger.warning(f'Could not connect to {addr_info[4]}: {str(e)}')
        else:
            # Did not break in for-loop above
            raise ScpiTransportException(f'Connection to host {address[0]} could not be established.')

    def close(self) -> None:
        # Do regular closing
        super().close()

        # Close socket as well
        self._sock.shutdown(socket.SHUT_RDWR)
        self._sock.close()
        self._sock = None



from .scpi_device import ScpiDevice
from .scpi_transport_serial import ScpiSerialTransport
from .scpi_transport_tcpip import ScpiTcpIpTransport
from .scpi_transport_usbtmc import ScpiUsbTmcTransport
from .scpi_resource import ScpiResource
from .scpi_types import *



from typing import NamedTuple, Any, Optional, Type, List
from .scpi_transport_base import ScpiTransports, ScpiTransportBase
import logging

_logger = logging.getLogger(__name__)

class ScpiResource(NamedTuple):
    transport: Type[ScpiTransports]
    address: Any
    location: Optional[str] = None
    name: Optional[str] = None
    manufacturer: Optional[str] = None
    model: Optional[str] = None
    serialnum: Optional[str] = None
    info: Optional[Any] = None

    @classmethod
    def discover(cls, transports: Optional[List[Type[ScpiTransports]]] = None, **kwargs) -> List['ScpiResource']:
        """Find SCPI devices connected to system.

        By default lists all available possible SCPI devices. However behaviour can be controlled using
        kwargs supported by the transport classes."""

        if transports is None: transports = ScpiTransportBase.__subclasses__()

        resources = []

        for transport in transports:
            # filter keyword arguments accepted by transport and call transport's find_devices method
            args = {k:v for k, v in kwargs.items() if k in transport.discover.__code__.co_varnames}
            argstr = ", ".join(["{}={}".format(k, v) for k,v in args.items()])

            try:
                new = transport.discover(**args)
                resources += new
                _logger.debug(
                    f'Running discover({argstr}) on {transport._transport_class}: {len(new)} resources found')
            except Exception as e:
                _logger.warning(
                    f'Exception during discover({argstr}) on {transport._transport_class}: {str(e)}')

        return resources

    @classmethod
    def from_resource_name(cls, resource_name: str, transports: Optional[List[Type[ScpiTransports]]] = None) -> 'ScpiResource':
        """Create resource object for a device using a VISA resource name"""
        if transports is None: transports = ScpiTransportBase.__subclasses__()

        for transport in transports:
            # Find a transport that accepts resource name
            try:
                resource = transport.from_resource_name(resource_name)
                if resource is not None:
                    _logger.debug(f"Resource name '{resource_name}' accepted by {transport._transport_class}")
                    break
            except Exception as e:
                # Some transports may not implement this method
                _logger.warning(
                    f"Exception during parsing of resource name '{resource_name}' by {transport._transport_class}: {str(e)}"
                )
        else:
            # No transport found that accepts resource name
            _logger.error(
                f"No transport found accepting resource name '{resource_name}'"
            )
            raise ValueError

        return resource

    @property
    def resource_class(self) -> str:
        return 'ScpiResource'

    @property
    def resource_type(self) -> str:
        return 'SCPI'

    @property
    def resource_info(self) -> str:
        return 'SCPI Resource'

    @property
    def resource_name(self) -> str:
        """Returns the VISA resource string identifying the device"""
        return self.transport.to_resource_name(self)